/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2009 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */


#ifndef OIL_PARSER_H
#define OIL_PARSER_H

#include <boost/spirit/core.hpp>
#include <boost/spirit/error_handling.hpp>
#include <boost/spirit/tree/ast.hpp>
#include <boost/spirit/utility/functor_parser.hpp>
#include <boost/spirit/utility/regex.hpp>
#include <vector>
#include <map>
#include <algorithm>

#include "toppers/text.hpp"
#include "toppers/oil/oil_object.hpp"



using namespace boost::spirit;

namespace toppers
{
	namespace oil
	{
		typedef std::vector<toppers::oil::oil_implementation::oil_object_impl*> oil_impl;
		typedef std::vector<toppers::oil::oil_definition::object_definition*> oil_def;
		
		class oil_parser
		{
			public:
				oil_parser(std::string* txt , std::vector<std::string> const& objlist)
				{
					description = txt;
					param_addable = true;
					m_object_list = objlist;
				}
				~oil_parser()
				{
					std::vector<toppers::oil::oil_implementation::oil_object_impl*>::iterator p;

					for(p = impl_list.begin() ; p != impl_list.end() ; p++)
					{
						delete *p;
					}

					std::vector<toppers::oil::oil_definition::object_definition*>::iterator q;

					for(q = def_list.begin() ; q != def_list.end() ; q++)
					{
						delete *q;
					}
				}
				int do_parse(oil_impl **impl , oil_def **def);
				int start_objimpl_analysis(std::string object_name);
				int start_objdef_analysis(std::string object_name_type);
				int add_parameter_objimpl(std::string parameter);
				int set_nextparam_name(std::string name);
				int add_attr_value(std::string str);
				int end_objimpl_analysis(void);
				int end_objdef_analysis(void);
				void enable_addparam(bool enabled);
				void dump_implementation(void);
				void dump_definition(void);
				void set_error(int position , std::string message);
				void get_error(int *position , std::string *message);

				std::string* get_description(void)
				{
					return description;
				}
				std::vector<std::string> m_object_list;
			protected:
				std::string *description;
				oil_impl impl_list;
				oil_def def_list;
			private:
				toppers::oil::oil_implementation::oil_object_impl* current_impl;
				std::string next_param_name;
				std::string attr_description;
				bool param_addable;
				toppers::oil::oil_definition::object_definition* current_def;
				std::string error_message;
				int error_position;
		};

		// @G[
		enum oil_grammer_error
		{
			open_brace , close_brace , semicolon , equal , keyword , comma
		};
		// @G[^`
		typedef assertion<oil_grammer_error> oil_assertion;
		typedef boost::spirit::guard<oil_grammer_error> oil_error_guard;
		
		// G[nh
		struct error_handler
		{
			oil_parser* container;
			error_handler(oil_parser* _container) : container(_container)
			{
			}
			template < class Scanner, class oil_grammer_error >
			boost::spirit::error_status<> operator()( Scanner const& scan, oil_grammer_error const& error ) const
			{
				std::string message;
				int pos;
				boost::spirit::error_status<> result =
					boost::spirit::error_status<>( boost::spirit::error_status<>::fail );

				// G[oʒu̎Zo
				pos = error.where - &((container->get_description()->c_str())[0]);
				
				// bZ[W̓o^
				switch ( error.descriptor )
				{
				case open_brace:
					message = "missing '{'";
					break;
				case close_brace:
					message = "missing '}'";
					break;
				case semicolon:
					message = "missing ';'";
					break;
				case equal:
					message = "missing '='";
					break;
				case keyword:
					message = "can't use keyword";
					break;
				case comma:
					message = "missing ','";
					break;
				}

				container->set_error(pos , message);
				return result;
			}
		};

		// t@N^
		struct match_objimpl_name
		{
			oil_parser* container;
			match_objimpl_name(oil_parser* _container) : container(_container)
			{
			}
			void operator()(char const *first, char const *last) const
			{
				std::string str(first , last);
				std::vector<std::string>::iterator p;
				bool hit = false;

				for(p = container->m_object_list.begin() ; p != container->m_object_list.end() ; p++)
				{
					if(*p == str)
					{
						hit = true;
						break;
					}
				}
				container->start_objimpl_analysis(str);
			}
		};

		struct find_implparam_name
		{
			oil_parser* container;
			find_implparam_name(oil_parser* _container) : container(_container)
			{
			}

			void operator()(char const *first, char const *last) const
			{
				string str(first , last);
				container->set_nextparam_name(str);
			}
		};

		struct match_impl_parameter
		{
			oil_parser* container;
			match_impl_parameter(oil_parser* _container) : container(_container)
			{
			}

			void operator()(char const *first, char const *last) const
			{
				string str(first , last);
				container->add_parameter_objimpl(str);
			}
		};
		
		// ̕ʂot@N^
		struct match_objimpl_end
		{
			oil_parser* container;
			match_objimpl_end(oil_parser* _container) : container(_container)
			{
			}

			void operator()(char val) const
			{
				container->end_objimpl_analysis();
			}
		};
		
		// IuWFNg`擪ot@N^
		struct match_objdef_name
		{
			oil_parser* container;
			match_objdef_name(oil_parser* _container) : container(_container)
			{
			}
			void operator()(char const *first, char const *last) const
			{
				string str(first , last);
				container->start_objdef_analysis(str);
			}
		};

		// p[^`ǉt@N^
		struct find_attrname
		{
			oil_parser* container;
			find_attrname(oil_parser* _container) : container(_container)
			{
			}
			void operator()(char const *first, char const *last) const
			{
				string str(first , last);
				container->set_nextparam_name(str);
			}
		};

		// IuWFNg`I[ot@N^
		struct match_objdef_end
		{
			oil_parser* container;
			match_objdef_end(oil_parser* _container) : container(_container){}

			void operator()(char val) const
			{
				container->end_objdef_analysis();
			}
		};
		
		// `̃p[^^Cvʂt@N^
		struct find_attribute
		{
			oil_parser* container;
			find_attribute(oil_parser* _container) : container(_container)
			{
			}
			void operator()(char const *first, char const *last) const
			{
				string str(first , last);
				container->add_attr_value(str);
			}
		};

		// `̃p[^^Cvʂ(_)
		struct find_attribute_float
		{
			oil_parser* container;
			find_attribute_float(oil_parser* _container) : container(_container)
			{
			}
			void operator()(double val) const
			{
				char buf[256];
				sprintf_s(buf , "%lf" , val);
				string str(buf);
				
				container->add_attr_value(str);
			}
		};


		
		// ENUM̃p[^邽߂̃t@N^
		struct enum_start
		{
			oil_parser* container;
			enum_start(oil_parser* _container) : container(_container){}

			void operator()(char val) const
			{
				container->enable_addparam(false);
			}
		};

		// ENUM̃p[^邽߂̃t@N^
		struct enum_end
		{
			oil_parser* container;
			enum_end(oil_parser* _container) : container(_container){}

			void operator()(char val) const
			{
				container->enable_addparam(true);
			}
		};
		
		// OILp[TNX
		struct oil_grammer :  public grammar<oil_grammer>
		{
			oil_parser* m_container;
			oil_grammer(oil_parser* container)
			{
				m_container = container;
			}
			template <typename T> struct definition
			{
				// @G[ޒ`
				oil_assertion assertion_open_brace , assertion_close_brace , assertion_semicolon;
				oil_assertion assertion_equal , assertion_keyword , assertion_comma;
				// G[nhO`
				oil_error_guard guard_implementation , guard_object , guard_object_def;

				// BNF
				rule<T> file;
				rule<T> oil_version , implementation_definition , application_definition;
				rule<T> implementation_spec_list , implementation_spec , implementation_def;
				rule<T> object , object_ref_type , object_name , object_temp;
				
				rule<T> name , null_str , boolean , string_literal;
				rule<T> multiple_specifier , auto_specifier;
				rule<T> number_range , number_list , number , float_range , default_number , default_float;
				rule<T> attribute_name;
				rule<T> default_string , default_bool , default_name;
				rule<T> bool_values;
				rule<T> enumerator , enumeration , enumerator_list;
				
				rule<T> impl_parameter_list;
				
				rule<T> uint32_impl_attr_def , int32_impl_attr_def , uint64_impl_attr_def , int64_impl_attr_def;
				rule<T> float_impl_attr_def , string_impl_attr_def , bool_impl_attr_def , enum_impl_attr_def;
				rule<T> reference_name;
				rule<T> impl_attr_def , impl_ref_def;
				
				rule<T> object_definition;
				rule<T> parameter , attribute_value;

				definition(oil_grammer const& self) :
					assertion_open_brace(open_brace) ,
					assertion_close_brace(close_brace) ,
					assertion_semicolon(semicolon) ,
					assertion_equal(equal) ,
					assertion_keyword(keyword) ,
					assertion_comma(comma)
				{
					// L֌W
					null_str = str_p("");
					name = (alpha_p | ch_p('_')) >> +(alnum_p | ch_p('_'));
					string_literal = ch_p('"') >> name >> ch_p('"');
					multiple_specifier = !(ch_p('[') >> ch_p(']'));
					auto_specifier = !str_p("WITH_AUTO");
					number = (str_p("0x") >> +xdigit_p) | (str_p("0X") >> xdigit_p) | int_p | uint_p;
					boolean = str_p("TRUE") | str_p("FALSE");
					
					// WAftHgl
					number_list = ch_p('[') >> number >> *(ch_p(',') >> number) >> ch_p(']');
					number_range = !((ch_p('[') >> number >> str_p("..") >> number >> ch_p(']'))
									| number_list);
					default_number = !((ch_p('=') >> number)
										| (ch_p('=') >> str_p("NO_DEFAULT"))
										| (ch_p('=') >> str_p("AUTO")));
					float_range = !(ch_p('[') >> real_p >> str_p("..") >> real_p >> ch_p(']'));
					default_float = !((ch_p('=') >> real_p)
										| (ch_p('=') >> str_p("NO_DEFAULT"))
										| (ch_p('=') >> str_p("AUTO")));
					
					enumerator = name >> !(ch_p('{') >> implementation_def >> ch_p('}'));
					enumerator_list = enumerator >> *(ch_p(',') >> enumerator);
					enumeration = ch_p('[')[enum_start(self.m_container)] >> 
										enumerator_list >> ch_p(']')[enum_end(self.m_container)];

					default_name = !((ch_p('=') >> name)
										| (ch_p('=') >> str_p("NO_DEFAULT"))
										| (ch_p('=') >> str_p("AUTO")));
					
					attribute_name = name | object;
					attribute_value = (boolean >> !(ch_p('{')[enum_start(self.m_container)] >>
											*parameter >> (ch_p('}')[enum_end(self.m_container)])))
											[find_attribute(self.m_container)]
										| number[find_attribute(self.m_container)]
										| real_p[find_attribute_float(self.m_container)]
										| string_literal[find_attribute(self.m_container)]
										// AUTO͉̒`őp\
										| (name >> !(ch_p('{')[enum_start(self.m_container)] >>
											*parameter >> ch_p('}')[enum_end(self.m_container)]))
											[find_attribute(self.m_container)];

					default_string = !((ch_p('=') >> ch_p('"') >> name >> ch_p('"'))
										| (ch_p('=') >> str_p("NO_DEFAULT"))
										| (ch_p('=') >> str_p("AUTO")));
					default_bool = !((ch_p('=') >> boolean)
										| (ch_p('=') >> str_p("NO_DEFAULT"))
										| (ch_p('=') >> str_p("AUTO")));
					bool_values = !(ch_p('[')[enum_start(self.m_container)] >>
											str_p("TRUE") >> !(ch_p('{') >> implementation_def >> ch_p('}')) >>
									ch_p(',') >>
											str_p("FALSE") >> !(ch_p('{') >> implementation_def >> ch_p('}')) >>
									ch_p(']')[enum_end(self.m_container)]);

					std::vector<std::string>::iterator p;
					std::string objtype = "(" , objref = "(";
					
					// K\̃Xg쐬
					for(p = self.m_container->m_object_list.begin() ; p != self.m_container->m_object_list.end() ; p++)
					{
						objtype += (*p) + "|";
						objref += (*p) + "_TYPE" + "|";
					}
					
					objtype[objtype.size() - 1] = ')';
					objref[objref.size() - 1] = ')';

					object = regex_p(objtype.c_str());
					object_ref_type = regex_p(objref.c_str());
					
					// IuWFNg`
					parameter = attribute_name[find_attrname(self.m_container)] >>
								assertion_equal(ch_p("=")) >> attribute_value >>
								assertion_semicolon(ch_p(";"));
					object_name = (object >> name)[match_objdef_name(self.m_container)];
					object_definition = guard_object(
											object_name >> assertion_open_brace(ch_p('{')) >> 
											*parameter >> assertion_close_brace(ch_p('}')) >> 
											assertion_semicolon(ch_p(';'))[match_objdef_end(self.m_container)])
											[error_handler(self.m_container)];

					// ^ʂ̃p[^wq
					uint32_impl_attr_def = str_p("UINT32") >> auto_specifier >> number_range >>
										attribute_name[find_implparam_name(self.m_container)] >> 
										multiple_specifier >> default_number >> 
										assertion_semicolon(ch_p(';'));
					int32_impl_attr_def = str_p("INT32") >> auto_specifier  >> number_range >> 
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_number >>
										assertion_semicolon(ch_p(';'));
					uint64_impl_attr_def = str_p("UINT64") >> auto_specifier  >> number_range >> 
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_number >>
										assertion_semicolon(ch_p(';'));
					int64_impl_attr_def = str_p("INT64") >> auto_specifier  >> number_range >> 
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_number >>
										assertion_semicolon(ch_p(';'));
					float_impl_attr_def = str_p("FLOAT") >> auto_specifier  >> float_range >> 
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_float >>
										assertion_semicolon(ch_p(';'));
					string_impl_attr_def = str_p("STRING") >> auto_specifier >> 
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_string >>
										assertion_semicolon(ch_p(';'));
					bool_impl_attr_def = str_p("BOOLEAN") >>
										auto_specifier >> bool_values >>
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_bool >>
										assertion_semicolon(ch_p(';'));
					enum_impl_attr_def = str_p("ENUM") >>
										auto_specifier >> enumeration >>
										attribute_name[find_implparam_name(self.m_container)] >>
										multiple_specifier >> default_name >>
										assertion_semicolon(ch_p(';'));

					// l
					impl_attr_def = uint32_impl_attr_def | int32_impl_attr_def
									| uint64_impl_attr_def | int64_impl_attr_def
									| float_impl_attr_def | string_impl_attr_def
									| bool_impl_attr_def | enum_impl_attr_def;
					
					reference_name = name | object;
					impl_ref_def = object_ref_type >> reference_name[find_implparam_name(self.m_container)] >>
									multiple_specifier >> assertion_semicolon(ch_p(';'));
					
					// p[^̌`
					implementation_def = *((impl_attr_def | impl_ref_def)
												[match_impl_parameter(self.m_container)]);

					// {̒`
					application_definition = 
							str_p("CPU") >> name >> ch_p('{') >> 
							*object_definition >> ch_p('}') >> ch_p(';');

					// 傫ȒPʂ̕@
					implementation_spec = 
						guard_object(object[match_objimpl_name(self.m_container)] >>
										assertion_open_brace(ch_p('{')) >>
										implementation_def >>
										assertion_close_brace(ch_p('}')) >>
										assertion_semicolon(ch_p(';')[match_objimpl_end(self.m_container)]))
										[error_handler(self.m_container)];
					implementation_spec_list = *implementation_spec;
					
					implementation_definition = "IMPLEMENTATION" >> name >> ch_p('{')
						>> implementation_spec_list >> ch_p('}') >> ch_p(';');

					oil_version = str_p("OIL_VERSION") 
								>> ch_p('=') >> ch_p('\"') >> real_p >> ch_p('\"') >> ch_p(';');	//
					file = oil_version >> implementation_definition >> application_definition >> *space_p;
				};
				rule<T> const& start() const
				{
					return file;                    // `[Ԃ
				}
			};
			
		};


	}
}

#endif	/* OIL_PARSER_H */
